#include "../SDK/foobar2000.h"
#include "../helpers/helpers.h"
#include "resource.h"
#include <math.h>
#include <commctrl.h>

DECLARE_COMPONENT_VERSION("VU meter", "0.2.0", 0);

//  in UTF-8
#define PLUS_MINUS_UTF8 "\xC2\xB1"

static HWND g_wnd;

static cfg_int cfg_enabled("enabled",0);

static cfg_struct_t<WINDOWPLACEMENT> cfg_window_placement("window_placement", 0);

struct vu_chunk
{
	double duration;
	double peak;
	double rms2; // squared, so it's only MS instead of RMS
	double avg;
};

enum {
	BAR_RANGE=1000,
};

static const double DB_MIN = -60.0;
static const double DB_MAX = +10.0;
static const double DB_RANGE = DB_MAX-DB_MIN;

static const double WINDOW_SIZE = 0.1; // actual value may differ

double ratio2dB(double x)
{
	static const double f = 20.0 / log(10);
	return log(x) * f;
}

int dB2bar(double x)
{
	double y = x-DB_MIN;
	int v = (int)(y / DB_RANGE * BAR_RANGE);
	return max(0,min(BAR_RANGE,v));
}

int ratio2bar(double x)
{
	return dB2bar(ratio2dB(x));
}

void set_vu_bar(HWND wnd, double dB)
{
	COLORREF color = dB>0 ? RGB(255,0,0) : (dB>-6 ? RGB(255,255,0) : RGB(0,255,0));
	uSendMessage(wnd, PBM_SETBARCOLOR, 0, color);
	uSendMessage(wnd, PBM_SETPOS, dB2bar(dB), 0);
}

void set_vu_text(HWND wnd, double dB)
{
	char buffer[32];
	const char * sign;
	if (dB < 0)
		sign = "-";
	else if (dB > 0)
		sign = "+";
	else
		sign = PLUS_MINUS_UTF8;
	int v = (int)(10*fabs(dB));
	sprintf(buffer, "%s%d.%ddB", sign, v/10, v%10);
	uSetWindowText(wnd, buffer);
}

static const dialog_resize_helper::param g_resize_helper_table[] = {
	{IDC_PEAKBAR,dialog_resize_helper::X_SIZE},
	{IDC_RMSBAR,dialog_resize_helper::X_SIZE},
	{IDC_DBMAX,dialog_resize_helper::X_MOVE},
};

class dialog_vu_meter : public dialog_helper::dialog
{
	dialog_resize_helper resize_helper;
	int height;

protected:
	virtual BOOL on_message(UINT msg, WPARAM wp, LPARAM lp)
	{
		// we'll have to handle WM_GETMINMAXINFO ourselves...
		if (!resize_helper.process_message(get_wnd(),msg,wp,lp) || msg == WM_GETMINMAXINFO)
		switch (msg)
		{
		case WM_INITDIALOG:
			{
				g_wnd = get_wnd();
				COLORREF color_bk = RGB(0,0,0);

				// set always on top
				SetWindowPos(get_wnd(), HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE|SWP_NOSIZE);

				// set dialog caption
				const char * core_version = core_version_info::get_version_string();
				if (!core_version && !*core_version) core_version = "foobar2000";
				uSetWindowText(get_wnd(), string_printf("VU meter [%s]", core_version));

				// get height for WM_GETMINMAXINFO
				RECT rc;
				GetWindowRect(get_wnd(), &rc);
				height = rc.bottom-rc.top;

				// set up bars
				HWND bar;
				bar = GetDlgItem(get_wnd(), IDC_PEAKBAR);
				uSendMessage(bar, PBM_SETRANGE, 0, MAKELPARAM(0, BAR_RANGE));
				uSendMessage(bar, PBM_SETBKCOLOR, 0, color_bk);
				bar = GetDlgItem(get_wnd(), IDC_RMSBAR);
				uSendMessage(bar, PBM_SETRANGE, 0, MAKELPARAM(0, BAR_RANGE));
				uSendMessage(bar, PBM_SETBKCOLOR, 0, color_bk);

				set_vu_text(GetDlgItem(get_wnd(), IDC_DBMIN), DB_MIN);
				set_vu_text(GetDlgItem(get_wnd(), IDC_DBMAX), DB_MAX);

				// set up textual dB display
				uSetDlgItemText(get_wnd(), IDC_PEAK, "");
				uSetDlgItemText(get_wnd(), IDC_RMS, "");

				// register modeless dialog
				modeless_dialog_manager::add(get_wnd());

				// restore placement
				apply_window_placement(cfg_window_placement, get_wnd());
			}
			break;
		case WM_DESTROY:
			read_window_placement(cfg_window_placement, get_wnd());
			modeless_dialog_manager::remove(get_wnd());
			g_wnd = 0;
			break;
		case WM_CLOSE:
			cfg_enabled = false;
			DestroyWindow(get_wnd());
			break;
		case WM_GETDLGCODE:
			return DLGC_WANTALLKEYS;
			break;
		case WM_GETMINMAXINFO:
			{
				LPMINMAXINFO info = (LPMINMAXINFO) lp;
				info->ptMaxTrackSize.y = height;
				info->ptMinTrackSize.x = 200;
				info->ptMinTrackSize.y = height;
			}
			break;
		case WM_COMMAND:
			switch (wp)
			{
			case IDCANCEL:
				cfg_enabled = false;
				DestroyWindow(get_wnd());
				break;
			}
			break;
		}
		return 0;
	}
public:
	dialog_vu_meter() : resize_helper(g_resize_helper_table,tabsize(g_resize_helper_table),0,0,0,0) {}
	inline void show()
	{
		run_modeless(IDD_VUMETER, 0);
	}
	inline void hide()
	{
		if (g_wnd) DestroyWindow(g_wnd);
	}
};

static dialog_vu_meter g_dialog;

class visualization_vu_meter : public visualization
{
	mem_block_list_t<vu_chunk> chunks;
	double total_duration;
	const double max_duration;
public:
	visualization_vu_meter() : total_duration(0.0), max_duration(WINDOW_SIZE) {}
	~visualization_vu_meter() {chunks.remove_all();}
	virtual bool is_active() {return !!cfg_enabled;}
	virtual bool need_spectrum() {return false;}

	virtual void on_data(const vis_chunk * data)
	{
		if (!cfg_enabled || data->flags & vis_chunk::FLAG_LAGGED) return;

		unsigned ch, nch = data->nch;
		unsigned n, m = data->samples;
		vu_chunk chunk;
		chunk.duration = data->get_duration();
		chunk.peak = 0.0;
		chunk.rms2 = 0.0;
		chunk.avg = 0.0;
		if (m > 0 && chunk.duration > 0)
		{
			vis_sample * ptr = data->data;
			for (n = 0; n < m; n++)
			{
				double cur = 0.0;
				for (ch = 0; ch < nch; ch++)
				{
					double cur2 = fabs(*ptr++);
					cur += cur2;
					if (chunk.peak < cur2) chunk.peak = cur2;
				}
				cur = cur / (double)nch;
				chunk.avg += cur;
				chunk.rms2 += cur*cur;
			}
			chunk.avg *= chunk.duration / (double)data->samples;
			chunk.rms2 *= chunk.duration / (double)data->samples;

			while (chunks.get_count() > 0 && total_duration + chunk.duration > max_duration)
			{
				total_duration -= chunks[0].duration;
				chunks.remove_by_idx(0);
			}
			chunks.add_item(chunk);
			total_duration += chunk.duration;

			if (!g_wnd)
			{
				g_dialog.show();
			}
			if (g_wnd)
			{
				double peak = 0.0;
				double rms = 0.0;
				double avg = 0.0;

				m = chunks.get_count();
				for (n = 0; n < m; n++)
				{
					if (peak < chunks[n].peak) peak = chunks[n].peak;
					rms += chunks[n].rms2;
					avg += chunks[n].avg;
				}
				avg /= total_duration;
				rms = sqrt(rms / total_duration);

				peak = ratio2dB(peak);
				rms = ratio2dB(rms);
				avg = ratio2dB(avg);

				set_vu_text(GetDlgItem(g_wnd, IDC_PEAK), peak);
				set_vu_text(GetDlgItem(g_wnd, IDC_RMS), rms);

				set_vu_bar(GetDlgItem(g_wnd, IDC_PEAKBAR), peak);
				set_vu_bar(GetDlgItem(g_wnd, IDC_RMSBAR), rms);
			}
		}

	}
	virtual void on_flush()
	{
		if (g_wnd)
		{
			uSetDlgItemText(g_wnd, IDC_PEAK, "");
			uSetDlgItemText(g_wnd, IDC_RMS, "");
			uSendDlgItemMessage(g_wnd, IDC_PEAKBAR, PBM_SETPOS, 0, 0);
			uSendDlgItemMessage(g_wnd, IDC_RMSBAR, PBM_SETPOS, 0, 0);
			total_duration = 0.0;
			chunks.remove_all();
		}
	}
};

static visualization_factory<visualization_vu_meter> foo;

class menu_item_vu_meter : public menu_item_main {
protected:
	virtual bool get_display_data(unsigned n,const ptr_list_base<metadb_handle> & data,string_base & out,unsigned & displayflags) {
		assert(n>=0 && n<get_num_items());
		if (!is_available(n)) return false;
		string8 temp;
		enum_item(n,temp);
		const char * ptr = strrchr(temp,'/');
		if (ptr) ptr++;
		else ptr = temp;
		out.set_string(ptr);
		displayflags = (is_checked(n) ? FLAG_CHECKED : 0) | (is_disabled(n) ? FLAG_DISABLED_GRAYED : 0);
		return true;
	}

	virtual unsigned get_num_items() {
		return 1;
	}

	virtual void enum_item(unsigned n, string_base & out) {
		if (n == 0) out = "Components/Visualization/VU meter";
	}

	virtual bool get_description(unsigned n, string_base & out) {
		if (n == 0) {
			if (!visualization::is_vis_manager_present())
				out = "Shows VU meter visualization (requires visualization manager to be installed).";
			else if (cfg_enabled)
				out = "Hides VU meter visualization.";
			else
				out = "Shows VU meter visualization.";
			return true;
		}
		return false;
	}

	virtual bool is_disabled(unsigned n) {
		if (n == 0) return !visualization::is_vis_manager_present();
		return true;
	}

	virtual void perform_command(unsigned n) {
		if (n == 0 && core_api::assert_main_thread()) {
			cfg_enabled = !cfg_enabled;
			if (!cfg_enabled)
				g_dialog.hide();
			else {
				g_dialog.show();
				if (!visualization::is_vis_manager_present())
					console::error("Visualization manager not present, please reinstall.");
			}
		}
	}

	virtual bool is_checked(unsigned n) {
		if (n == 0) return !!cfg_enabled;
		return false;
	};
};

static menu_item_factory< menu_item_vu_meter > foo2;

class initquit_vu_meter : public initquit
{
public:
	virtual void on_init()
	{
		if (cfg_enabled) g_dialog.show();
	}
	virtual void on_quit()
	{
		if (g_wnd) g_dialog.hide();
	}
};

static initquit_factory< initquit_vu_meter > foo3;
